use std::{error::Error, sync::Arc, time::Instant};

use crate::stations_list::{Arrivals, Departures, FilterTransports, StationsList};

pub struct Game {
    stations: Arc<StationsList>,
    history: Vec<(String, Option<String>)>,
    location: String,
    location_name: Option<String>,
    departures: Result<Departures, Box<dyn Error>>,
    target: String,
    target_name: Option<String>,
    location_activity_count: usize,
    target_activity_count: usize,
    filter_transports: FilterTransports,
    hint: Option<(Arrivals, Instant)>,
}
impl Game {
    pub fn new(
        filter_transports: FilterTransports,
        location: Option<String>,
        target: Option<String>,
        stations: Arc<StationsList>,
    ) -> Result<Self, Box<dyn Error>> {
        let location = if let Some(location) = location {
            match stations
                .find_stations(&location)
                .first()
                .and_then(|(id, _)| {
                    stations.get_station(id, |station| {
                        station
                            .query_activity_count(id, false, 10, filter_transports.only_known())
                            .map(|c| (id.to_owned(), c))
                    })
                }) {
                Some(Ok((id, c))) => (c, id),
                Some(Err(e)) => Err(e)?,
                None => Err(format!("start does not exist"))?,
            }
        } else {
            let mut location: Option<(usize, String)> = None;
            let mut location_counter = 0;
            for _ in 0..10 {
                if let Some(st) = stations.get_random_station(3, |id, station| {
                    if let Some(act) = station
                        .query_activity_count(id, false, 10, filter_transports.only_known())
                        .ok()
                        .filter(|v| *v > 1)
                    {
                        Some((act, id.to_owned()))
                    } else {
                        None
                    }
                }) {
                    location_counter += 1;
                    if location.as_ref().is_none_or(|p| p.0 < st.0) {
                        location = Some(st);
                    }
                    // after 5 locations with 2+ departures, take the best one
                    if location_counter >= 5 {
                        break;
                    }
                }
            }
            location.ok_or("could not find a start location")?
        };
        let target = if let Some(target) = target {
            match stations.find_stations(&target).first().and_then(|(id, _)| {
                stations.get_station(id, |station| {
                    station
                        .query_activity_count(id, false, 10, filter_transports.only_known())
                        .map(|c| (id.to_owned(), c))
                })
            }) {
                Some(Ok((id, c))) => (c, id),
                Some(Err(e)) => Err(e)?,
                None => Err(format!("ziel does not exist"))?,
            }
        } else {
            let mut target: Option<(usize, String)> = None;
            let mut target_counter = 0;
            for _ in 0..10 {
                if let Some(st) = stations.get_random_station(3, |id, station| {
                    if let Some(act) = (id != location.1.as_str())
                        .then(|| {
                            station
                                .query_activity_count(id, true, 10, filter_transports.only_known())
                                .ok()
                                .filter(|v| *v > 1)
                        })
                        .flatten()
                    {
                        Some((act, id.to_owned()))
                    } else {
                        None
                    }
                }) {
                    target_counter += 1;
                    if target.as_ref().is_none_or(|p| p.0 < st.0) {
                        target = Some(st);
                    }
                    // after 5 targets with 2+ arrivals, take the best one
                    if target_counter >= 5 {
                        break;
                    }
                }
            }
            target.ok_or("could not find a target station")?
        };
        Ok(Self {
            stations,
            history: vec![],
            location: location.1,
            location_name: None,
            departures: Err("".into()),
            target: target.1,
            target_name: None,
            location_activity_count: location.0,
            target_activity_count: target.0,
            filter_transports,
            hint: None,
        })
    }

    pub fn history(&self) -> &[(String, Option<String>)] {
        &self.history
    }
    pub fn location(&self) -> &str {
        &self.location
    }
    pub fn location_name(&mut self) -> &str {
        if self.location_name.is_none() {
            self.location_name = self
                .stations
                .get_station(self.location(), |s| s.name().to_owned());
        }
        self.location_name
            .as_ref()
            .map(|v| v.as_str())
            .unwrap_or(self.location())
    }
    pub fn location_name_immut(&self) -> &str {
        self.location_name
            .as_ref()
            .map(|v| v.as_str())
            .unwrap_or(self.location())
    }
    pub fn target_name_immut(&self) -> &str {
        self.target_name
            .as_ref()
            .map(|v| v.as_str())
            .unwrap_or(self.target())
    }
    pub fn target(&self) -> &str {
        &self.target
    }
    pub fn target_name(&mut self) -> &str {
        if self.target_name.is_none() {
            self.target_name = self
                .stations
                .get_station(self.target(), |s| s.name().to_owned());
        }
        self.target_name
            .as_ref()
            .map(|v| v.as_str())
            .unwrap_or(self.target())
    }
    pub fn location_activity_count(&self) -> usize {
        self.location_activity_count
    }
    pub fn target_activity_count(&self) -> usize {
        self.target_activity_count
    }

    pub fn update_departures(&mut self) {
        let v = std::mem::replace(&mut self.departures, Err("".into())).or_else(|_| {
            match self.stations.get_station(&self.location, |location| {
                location.query_departures(&self.location, 15, self.filter_transports)
            }) {
                Some(v) => {
                    if let Ok(v) = &v {
                        self.stations.add_new_from_departures(v);
                    }
                    v
                }
                None => Err(
                    "current location no longer exists (this is most likely a server error)".into(),
                ),
            }
        });
        self.departures = v;
    }
    pub fn departures_cached(&self) -> Result<&Departures, &Box<dyn Error>> {
        self.departures.as_ref()
    }
    /// Err if check failed, Ok(false) if 0 or 1 departures, Ok(true) otherwise
    pub fn go_to_check(&self, location: &str) -> Result<usize, Box<dyn Error>> {
        self.stations
            .requery_station_if_necessary_or_add_new(location);
        if let Some(act) = self.stations.get_station(location, |s| {
            s.query_activity_count(location, false, 10, self.filter_transports.only_known())
        }) {
            Ok(act?)
        } else {
            Err(format!("no such station ({location})").into())
        }
    }

    /// true if win
    pub fn go_to_unchecked(&mut self, location: String) -> bool {
        let ploc = std::mem::replace(&mut self.location, location);
        let plocn = self.location_name.take();
        self.history.push((ploc, plocn));
        self.departures = Err("".into());
        self.location == self.target
    }

    pub fn get_hint(&mut self) -> Result<&Arrivals, Box<dyn Error>> {
        if self
            .hint
            .as_ref()
            .is_none_or(|v| v.1.elapsed().as_secs_f64() > 60.0)
        {
            let o = self
                .stations
                .get_station(&self.target, |station| {
                    station.query_arrivals(&self.target, 15, self.filter_transports)
                })
                .ok_or_else(|| "failed to find target station");
            if let Ok(Ok(o)) = o {
                self.hint = Some((o, Instant::now()));
            } else if self.hint.is_none() {
                self.hint = Some((o??, Instant::now()));
            }
        }
        Ok(&self.hint.as_ref().unwrap().0)
    }
}
